#include <fstream.h>
#include <ctype.h>
#include <String.h>
#include <assert.h>

#include <MAPLAB/iterators.h>
#include <MAPLAB/IO/information_base.h>

#if (FileReader_DEBUG || FILEREADER_DEBUG || DEBUG)
#include <iostream.h>
#define File_debug(command) command
#else
#define File_debug(command)
#endif

template <class CoordType, class NameType>
class File_reader : public Information_base<CoordType, NameType>
{
public:
  File_reader(const char* file_name)
  : filename(""), creator(""), date(""), width(0), height(0),
    label_positions(1), font("Roman"), font_size(12),
    bounding_box_set(false), contains_name_candidates(false),
    window_set(false), x_scale(1), y_scale(1), x_min_set(false),
    y_min_set(false), x_max_set(false), y_max_set(false)
  {
    scan_file(file_name);
    insert_name_candidates();
  }
  String& get_filename() { return filename; }
  
  String& get_creator() { return creator; }
  
  String& get_date() { return date; }
  
  String& get_fontname() { return font; }
  
  int get_fontsize() { return font_size; }
  
  double get_xscale() { return x_scale; }
  
  double get_yscale() { return y_scale; }    
private:
  void skip_leading_spaces(String& str)
  {
    while (!str.empty() && str.firstchar() == ' ')
      str = str.after(' ');
  }
  
  value_type get_next_num(String& str)
  {
    skip_leading_spaces(str);
    String sub_str = str.through(RXdouble);
    str = str.after(sub_str);
    return (value_type) atof(sub_str.chars());
  }
  void replace_umlauts(String& str)
  {
    str.gsub("", "\344");
    str.gsub("", "\304");
    str.gsub("", "\366");
    str.gsub("", "\326");
    str.gsub("", "\374");
    str.gsub("", "\334");
    str.gsub("", "\337");
  }
  
  String filename;
  String creator;
  String date;
  value_type width;
  value_type height;
  int label_positions;
  String font;
  int font_size;
  bool bounding_box_set, contains_name_candidates, window_set;
  value_type box_llx, box_lly, box_urx, box_ury, window_llx, window_lly,
    window_urx, window_ury;
  value_type x_min, y_min, x_max, y_max;
  bool x_min_set, y_min_set, x_max_set, y_max_set;
  double x_scale, y_scale;
  void parse_candidates(value_type x, value_type y, String& str)
  {
    Site_it current = insert_site(x, y);
    char first_char;
    do
    {
      str = str.after('(');
      insert_label(current, get_next_num(str), get_next_num(str),
                   get_next_num(str), get_next_num(str));
      str = str.after(')');
      if (str.empty()) break;
      skip_leading_spaces(str);
      first_char = str.firstchar();
    } while (first_char == '(');
  }
  void string_to_candidates(value_type x, value_type y, String& str)
  {
    contains_name_candidates = true;
    String name = str.after('"');
    name = name.before('"');
    Site_it current = insert_site(x, y);
    replace_umlauts(name);
    set_name(current, name);
  }
  void scan_site_label_info(String& str)
  {
    value_type x_coord = get_next_num(str);
    value_type y_coord = get_next_num(str);
  
    if ((x_min_set && (x_min > x_coord)) ||
        (x_max_set && (x_max < x_coord)) ||
        (y_min_set && (y_min > y_coord)) ||
        (y_max_set && (y_max < y_coord)))
      return;
  
    skip_leading_spaces(str);
    if (str.empty() || (isdigit(str.firstchar())))
    {
      Site_it current = insert_site(x_coord, y_coord);
      value_type wdth, hght;
      if (str.empty()) { wdth = width; hght = height; }
      else { wdth = get_next_num(str); hght = get_next_num(str); }
      if (label_positions == 4)
        insert_four_labels(current, wdth, hght);
      else insert_label(current, x_coord, y_coord, x_coord + wdth,
                        y_coord + hght);
    }
    else if (str.firstchar() == '(') parse_candidates(x_coord, y_coord, str);
    else if (str.firstchar() == '"') string_to_candidates(x_coord, y_coord,
                                                           str);
  }
  void scan_global_variable_info(String& str)
  {
    String buff;
    if (str.contains("%Filename:"))
    {
      filename = str.after("%Filename:");
      skip_leading_spaces(filename);
    }
    else if (str.contains("%Creator:"))
    {
      creator = str.after("%Creator:");
      skip_leading_spaces(creator);
    }
    else if (str.contains("%Date:"))
    {
      date = str.after("%Date:");
      skip_leading_spaces(date);
    }
    else if (str.contains("%Labelsize:"))
    {
      buff = str.after("%Labelsize:");
      width = get_next_num(buff);
      height = get_next_num(buff);
    }
    else if (str.contains("%Labelpositions:"))
    {
      buff = str.after("%Labelpositions:");
      // label_positions = get_next_num(buff);
      label_positions = 1;
    }
    else if (str.contains("%Font:"))
    {
      buff = str.after("%Font:");
      skip_leading_spaces(buff);
      font = buff.before(' ');
      buff = buff.after(font);
      skip_leading_spaces(buff);
      if (!buff.empty()) font_size = get_next_num(buff);
    }
    else if (str.contains("%Boundingbox:"))
    {
      buff = str.after("%Boundingbox:");
      box_llx = get_next_num(buff);
      box_lly = get_next_num(buff);
      box_urx = get_next_num(buff);
      box_ury = get_next_num(buff);
      bounding_box_set = true;
    }
    else if (str.contains("%Window:"))
    {
      buff = str.after("%Window:");
      window_llx = get_next_num(buff);
      window_lly = get_next_num(buff);
      window_urx = get_next_num(buff);
      window_ury = get_next_num(buff);
      window_set = true;
    }
    else if (str.contains("%Xmin:"))
    {
      buff = str.after("%Xmin:");
      x_min = get_next_num(buff);
      x_min_set = true;
    }
    else if (str.contains("%Ymin:"))
    {
      buff = str.after("%Ymin:");
      y_min = get_next_num(buff);
      y_min_set = true;
    }
    else if (str.contains("%Xmax:"))
    {
      buff = str.after("%Xmax:");
      x_max = get_next_num(buff);
      x_max_set = true;
    }
    else if (str.contains("%Ymax:"))
    {
      buff = str.after("%Ymax:");
      y_max = get_next_num(buff);
      y_max_set = true;
    }
    else if (str.contains("%Xscale:"))
    {
      buff = str.after("%Xscale:");
      x_scale = get_next_num(buff);
    }
    else if (str.contains("%Yscale:"))
    {
      buff = str.after("%Yscale:");
      y_scale = get_next_num(buff);
    }
  }
  void scan_file(const char* file_name)
  {
    const int max = 1024;
    char buff[max], new_line;
  
    File_debug(cout << " File_reader: file_name = " << file_name << endl;)
    ifstream datafile(file_name);
    assert(datafile);
  
    while (datafile.get(buff, max))
    {
      String line(buff);
      skip_leading_spaces(line);
  
      if (!line.empty())
      {
        char first_char = line.firstchar();
        if (isdigit(first_char) || (first_char == '-'))
          scan_site_label_info(line);
        else if (first_char == '%')
          scan_global_variable_info(line);
      }
      datafile.get(new_line);
    }
    scale(x_scale, y_scale);
    if (contains_name_candidates)
      set_width_height(font, font_size);
    if (bounding_box_set)
      make_bounding_box(box_llx, box_lly, box_urx, box_ury);
    if (window_set)
    {
      make_window(window_llx, window_lly, window_urx, window_ury);
      make_bounding_box(box_llx, box_lly, box_urx, box_ury);
    }
    File_debug(cout << " File_reader: done." << endl;)
  }
};
