#ifndef STL_EXT
#define STL_EXT

// compiler dependencies
#if (__SUNPRO_CC)
#define typename
#endif

// standard libraries
#include <iostream.h>

// STL
#include <vector.h>
#include <iterator.h>

template <class InputIt, class Second_type>
class pair_second_it
{
public:
  typedef Second_type value_type;

  // constructors
  pair_second_it() {}
  pair_second_it(InputIt const &it): my_it(it) {}

  // public member functions
  pair_second_it &operator++ () { ++my_it; return *this;}
  Second_type const &operator* () { return (*my_it).second;}
  bool operator== (pair_second_it const &it) const
  { return my_it == it.my_it;}
  bool operator!= (pair_second_it const &it) const
  { return my_it != it.my_it;}

private:
  InputIt my_it;
};
template <class InputIt, class T, class Predicate>
class predicate_it
{
public:
  typedef T value_type;

  // constructors:
  // stopper
  predicate_it(): my_valid(false) {}
  // starter
  predicate_it(InputIt const &beg, InputIt const &end,
               Predicate pred = Predicate())
  : my_current(beg), my_end(end), my_pred(pred), my_valid(false)
  {
    if (beg != end) my_valid = my_pred(my_current);
  }

  // public member functions
  predicate_it &operator++ ()
  {
    ++my_current;
    if (my_current != my_end)
      my_valid = my_pred(my_current);
    else my_valid = false;
    return *this;
  }
  value_type const operator* () { return *my_current;}
  bool operator== (predicate_it const &it) const
  {
    if ((my_valid) && (it.my_valid))
      return (my_current == it.my_current);
    return my_valid == it.my_valid;
  }
  bool operator!= (predicate_it const &it) const
  {
    return !(*this == it);
  }

private:
  InputIt my_current, my_end;
  Predicate my_pred;
  bool my_valid;
};
template <class InputIt, class T, class Filter>
class filter_it
{
public:
  typedef T value_type;

  // constructors
  filter_it() {}
  // starter
  filter_it(InputIt const &beg, InputIt const &end,
            Filter filter = Filter())
  : my_current(beg), my_end(end), my_filter(filter)
  {
    advance();
  }
  // stopper
  filter_it(InputIt const &end): my_current(end), my_end(end)
  {}

  // public member functions
  filter_it &operator++ ()
  {
    my_current++;
    advance();
    return *this;
  }
  value_type operator* () { return *my_current;}
  bool operator== (filter_it const &it) const
  {
    return (my_current == it.my_current);
  }
  bool operator!= (filter_it const &it) const
  {
    return !(*this == it);
  }

private:
  InputIt my_current, my_end;
  Filter my_filter;

  void advance()
  {
    while ((my_current != my_end) && !my_filter(*my_current))
      my_current++;
  }
};
template <class InputIt, class State>
class state_it
{
public:
  typedef pair<InputIt, State> value_type;

  // constructors
  state_it()
  {}
  state_it(InputIt const &beg, State state = State())
  : my_current(beg), my_state(state)
  {}

  // public member functions
  state_it &operator++ ()
  {
    if (my_state(my_current))
      my_current++;
    ++my_state;
    return *this;
  }
  value_type const operator* () const
  { return value_type(my_current, my_state); }
  bool operator== (state_it const &it) const
  {
    return ((my_current == it.my_current) && (my_state == it.my_state));
  }
  bool operator!= (state_it const &it) const { return !(*this == it); }

private:
  InputIt my_current;
  State my_state;
};
template <class InputIt, class DA>
class member_access_it
{
public:
  typedef typename DA::value_type value_type;

  // constructors
  member_access_it()
  {}
  member_access_it(InputIt const &it, DA da = DA()): my_it(it), my_da(da)
  {}

  // public member functions
  member_access_it &operator++ ()
  {
    ++my_it;
    return *this;
  }
  value_type const operator* () { return get(my_da, *my_it);}
  bool operator== (member_access_it const &it) const
  {
    return my_it == it.my_it;
  }
  bool operator!= (member_access_it const &it) const
  {
    return !(my_it == it.my_it);
  }
  InputIt get_iterator()
  {
    return my_it;
  }
  DA get_da()
  {
    return my_da;
  }
private:
  InputIt my_it;
  DA my_da;
};
template <class InputIt1, class InputIt2>
class concat_it
{
  enum Iterator_state {One, Two};
public:
  typedef typename InputIt1::value_type value_type;

  // constructors
  concat_it()
  {}
  // starter
  concat_it(InputIt1 const &beg1, InputIt1 const &end1,
            InputIt2 const &beg2)
  : my_current1(beg1), my_end1(end1), my_current2(beg2), my_state(One)
  { if (my_current1 == my_end1) my_state = Two; }
  // stopper
  concat_it(InputIt2 const &end2): my_end2(end2), my_state(Two)
  {}

  // public member functions
  concat_it &operator++ ()
  {
    if (my_state == One)
    {
      ++my_current1;
      if (my_current1 == my_end1) my_state = Two;
    }
    else ++my_current2;
    return *this;
  }
  value_type const operator* ()
  {
    value_type vt;
    if (my_state == One) vt = *my_current1;
    else vt = *my_current2;
    return vt;
  }
  bool operator== (concat_it const &it) const
  {
    if (my_state == it.my_state)
    {
      if (my_state == One) return (my_current1 == it.my_current1);
      else return (my_current2 == it.my_current2);
    }
    return false;
  }
  bool operator!= (concat_it const &it) const { return !(*this == it); }

private:
  InputIt1 my_current1, my_end1;
  InputIt2 my_current2, my_end2;
  Iterator_state my_state;
};
#ifdef __KCC
using namespace std;
#endif

#if defined(__GNUG__) && __GNU_MINOR__ < 7
template <class T>
int *distance_type(istream_iterator<T> const &)
{
  return 0;
}
template <class T>
input_iterator_tag iterator_category(istream_iterator<T> const &)
{
  return input_iterator_tag();
}
#endif

template <class Iterator, class Distance>
Distance intern_lower_distance(Iterator &beg, Iterator &end, Distance *,
                               output_iterator_tag const &)
{
  Distance dist(0);
  distance(beg, end, dist);
  return dist;
}
template <class Iterator, class Distance>
Distance intern_lower_distance(Iterator &beg, Iterator &end, Distance *,
                               forward_iterator_tag const &)
{
  Distance dist(0);
  distance(beg, end, dist);
  return dist;
}
template <class Iterator, class Distance>
Distance intern_lower_distance(Iterator &beg, Iterator &end, Distance *,
                               bidirectional_iterator_tag const &)
{
  Distance dist(0);
  distance(beg, end, dist);
  return dist;
}
template <class Iterator, class Distance>
Distance intern_lower_distance(Iterator &beg, Iterator &end, Distance *,
                               random_access_iterator_tag const &)
{
  Distance dist(0);
  distance(beg, end, dist);
  return dist;
}

template <class T, class Distance>
Distance intern_lower_distance(istream_iterator<T, Distance> &beg,
                               istream_iterator<T, Distance> &end,
                               Distance *, input_iterator_tag const &)
{
  return 0;
}

// this is a temporary work-around for Iso_rectangle_intersection
template <class Iterator, class State>
int lower_distance(state_it<Iterator, State> begin,
                   state_it<Iterator, State> end)
{
  return (4*intern_lower_distance((*begin).first, (*end).first,
                                  distance_type((*begin).first),
                                  iterator_category((*begin).first)));
}

template <class Iterator>
int lower_distance(Iterator begin, Iterator const &end)
{
  // return intern_lower_distance(begin, end, distance_type(begin),
  //                             iterator_category(begin));
  int distance = 0;
  while (begin != end)
  {
    ++distance;
    ++begin;
  }
  return distance;
}

#endif
